%{
#include <stdio.h>
#define YYSTYPE char*
extern int yylex();
extern unsigned yylineno;

#include "libxml/tree.h"
extern xmlDocPtr document_xml ;
extern xmlNodePtr MetaGrammar ;
extern xmlNodePtr CurrentClass ;
extern int class_found ;
extern int error_symb ;
extern int keep_comments;

#include "hash.h"
#include "symbol.h"

//#define DEBUGYACC 0
#ifdef DEBUGYACC
#define DBUGYACC(x) x
#else
#define DBUGYACC(x)
#endif

#define MALLOCERROR(xxxx) {printf("malloc_error:\t%s (line number: %d)\n",xxxx,yylineno);exit(-1);}

char * cp_str (char * str1, char * str2) {
  char * new_str ;
  if (str1 == NULL) return (str2) ;
  if (str2 == NULL) return (str1) ;
  new_str = (char*) calloc(strlen(str1)+strlen(str2)+1,sizeof (char)) ;
  sprintf (new_str, "%s%s", str1, str2) ;
  free (str1); free(str2) ;
  return (new_str) ;
}

unsigned char * xmlstrdup (unsigned char * in) {
    unsigned char *out;
    int ret,size,out_size,temp;
    xmlCharEncodingHandlerPtr handler;
    
    size = (int)strlen(in)+1; 
    out_size = size*2-1; 
    out = malloc((size_t)out_size); 

    if (out) {
        handler = xmlFindCharEncodingHandler("iso-8859-1") ;
        if (!handler) {
            free (out);
            out = NULL;
	    printf ("xmlstrdup: no handler\n") ; exit (1) ;
        }
    }
    if (out) {
      temp=size-1;
      ret = handler->input(out, &out_size, in, &temp);
      if ((ret < 0) || (temp-size+1)) {
	if (ret < 0) {
	  printf ("xmlstrdup: conversion wasn't successful.\n");
	}
	else {
	  printf ("xmlstrdup: conversion wasn't successful. converted: %i octets.\n", temp);
	}
	free (out);
	out = NULL;
	exit (1) ;
      }
      else {
	out = realloc(out,out_size+1); 
	out[out_size]=0; /*null terminating out*/
      }
    }
    else {
      MALLOCERROR ("xmlstrdup") ;
    }
    return (out);
}
 
 
%}

%token CLASS DOCUMENTATION
%token IDENT MACRONAME TEMPLATE PATH DISABLE
%token LCURLYBK RCURLYBK LSQBK RSQBK LPAREN RPAREN
%token INHERIT FATHERREL
%token MINUS PLUS XPLUS STAR NEG PREC ANCESTOR
%token COMMA COLON SEMICOLON DOUBLECOLON EQUAL XEQUAL DOT ALT JOIN BANG IMP EQUIV
%token DESC NODE OPT VALUE VAR FATHER
%left ALT
%left JOIN
%start Axiom
%%
Axiom: Dcllist Doc
             {
	       char * comment_str = $2 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 xmlAddChild (MetaGrammar, comment_node) ;
	       }
	     } ;

Dcllist: Dcl {
             DBUGYACC(printf ("One Dcl\n") ;);
             }
       | Dcllist Dcl
       | error {printf ("\nLine number: %d\n",yylineno);} ;

Dcl: Doc CLASS IDENT
             {
	       class_found += 1;
	       CurrentClass = xmlNewChild (MetaGrammar, NULL, (xmlChar*)"class", NULL) ;
	       xmlSetProp (CurrentClass, (xmlChar*)"name", (xmlChar*)$3) ;
	       new_class_symb ($3, 1, yylineno) ;
	       char * comment_str = $1 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 xmlAddChild (CurrentClass, comment_node) ;
	       }
	     }
     Doc
             {
	       char * comment_str = $5 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 xmlAddChild (CurrentClass, comment_node) ;
	       }
	     }
     LCURLYBK Doc
             {
	       char * comment_str = $8 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 xmlAddChild (CurrentClass, comment_node) ;
	       }
	     }
     Proplist RCURLYBK
             {
	       /* reinitialiser pour une autre classe */
	       CurrentClass = NULL ;
             }
   | Doc TEMPLATE MACRONAME
             {
	       new_template_macro_symb ($3, 1, yylineno) ;
             }
     EQUAL Value
             {
	       char * comment_str = $1 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 xmlAddChild (MetaGrammar, comment_node) ;
	       }
	       xmlNodePtr templateNode = xmlNewNode (NULL, (xmlChar*)"vLib") ;
	       xmlSetProp (templateNode, (xmlChar*)"id", (xmlChar*)$3) ;
	       xmlAddChild (templateNode, (xmlNodePtr)$6) ;
	       xmlAddChild (MetaGrammar, templateNode) ;
             }
   | Doc PATH MACRONAME
             {
	       new_path_macro_symb ($3, 1, yylineno) ;
             }
     EQUAL Path
             {
	       char * comment_str = $1 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 xmlAddChild (MetaGrammar, comment_node) ;
	       }
	       xmlNodePtr MacroNode = xmlNewNode (NULL, (xmlChar*)"path") ;
	       xmlSetProp (MacroNode, (xmlChar*)"id", (xmlChar*)$3) ;
	       xmlAddChild (MacroNode, (xmlNodePtr)$6) ;
	       xmlAddChild (MetaGrammar, MacroNode) ;
             }
   | Doc DISABLE IDENT
             {
	       char * comment_str = $1 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 xmlAddChild (MetaGrammar, comment_node) ;
	       }
	       xmlNodePtr DisableNode = xmlNewNode (NULL, (xmlChar*)"disable") ;
	       xmlSetProp (DisableNode, (xmlChar*)"name", (xmlChar*)$3) ;
	       xmlAddChild (MetaGrammar, DisableNode) ;
             }
   | error RCURLYBK {printf ("\nLine number: %d\n",yylineno);} ;

Proplist: 
        | Property Doc
             {
	       char * comment_str = $2 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr property_node = (xmlNodePtr)$1 ;
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 DBUGYACC(printf ("Commentaire dans une propriete (1): %s comment_node=%o property_node=%o\n",
				  (xmlChar*)$2, comment_node, property_node););
		 xmlAddNextSibling (property_node, comment_node) ;
	       }
	     }
        | Property SEMICOLON Doc Proplist
             {
	       char * comment_str = $3 ;
	       if (comment_str && strlen (comment_str) > 0 && keep_comments) {
		 xmlNodePtr property_node = (xmlNodePtr)$1 ;
		 xmlNodePtr comment_node = xmlNewComment ((xmlChar*)comment_str) ;
		 DBUGYACC(printf ("Commentaire dans une propriete (liste): %s comment_node=%o property_node=%o Proplist=%o\n",
				  (xmlChar*)$3, comment_node, property_node, $4););
		 xmlAddNextSibling (property_node, comment_node) ;
	       }
	     }
        | error SEMICOLON {printf ("\nLine number: %d\n",yylineno);};

Property: Heritage
        | Pere
        | Ancetre
        | Precedence
        | Equality
        | Desc
        | Equation
        | Node
        | Needs
        | Provides
        | Guard
        | Atomic ;

Heritage: INHERIT IDENT
             {
	       DBUGYACC(printf ("Propriete Heritage: %s\n",$2););
	       xmlNodePtr InheritNode = xmlNewChild (CurrentClass, NULL, (xmlChar*)"super", NULL) ;
	       xmlSetProp (InheritNode, (xmlChar*)"name", (xmlChar*)$2) ;
	       new_class_symb ($2, 0, yylineno) ;
	       $$ = (xmlChar*) InheritNode ;
             } ;

Pere: Idname FATHERREL Idname
             {
               DBUGYACC(printf ("Propriete Pere: %s %s\n",$1,$3););
	       xmlNodePtr FatherNode = xmlNewChild (CurrentClass, NULL, (xmlChar*)"relation", NULL) ;
	       xmlSetProp (FatherNode, (xmlChar*)"rel", (xmlChar*)"father") ;
	       xmlSetProp (FatherNode, (xmlChar*)"arg1", (xmlChar*)$1) ;
	       xmlSetProp (FatherNode, (xmlChar*)"arg2", (xmlChar*)$3) ;
	       $$ = (xmlChar*) FatherNode ;
             } ;

Ancetre: Idname ANCESTOR Idname
             {
               DBUGYACC(printf ("Propriete Anctre: %s %s\n",$1,$3););
	       xmlNodePtr AncestorNode = xmlNewChild (CurrentClass, NULL, (xmlChar*)"relation", NULL) ;
	       xmlSetProp (AncestorNode, (xmlChar*)"rel", (xmlChar*)"dominates") ;
	       xmlSetProp (AncestorNode, (xmlChar*)"arg1", (xmlChar*)$1) ;
	       xmlSetProp (AncestorNode, (xmlChar*)"arg2", (xmlChar*)$3) ;
	       $$ = (xmlChar*) AncestorNode ;
             } ;

Precedence: Idname PREC Idname
             {
	       DBUGYACC(printf ("Propriete Prcdence: %s %s\n",$1,$3););
	       xmlNodePtr PrecNode = xmlNewChild (CurrentClass, NULL, (xmlChar*)"relation", NULL) ;
	       xmlSetProp (PrecNode, (xmlChar*)"rel", (xmlChar*)"precedes") ;
	       xmlSetProp (PrecNode, (xmlChar*)"arg1", (xmlChar*)$1) ;
	       xmlSetProp (PrecNode, (xmlChar*)"arg2", (xmlChar*)$3) ;
	       $$ = (xmlChar*) PrecNode ;
             } ;

Equality: Idname EQUAL Idname
             {
	       DBUGYACC(printf ("Propriete Equality: %s %s\n",$1,$3););
	       xmlNodePtr PrecNode = xmlNewChild (CurrentClass, NULL, (xmlChar*)"relation", NULL) ;
	       xmlSetProp (PrecNode, (xmlChar*)"rel", (xmlChar*)"equals") ;
	       xmlSetProp (PrecNode, (xmlChar*)"arg1", (xmlChar*)$1) ;
	       xmlSetProp (PrecNode, (xmlChar*)"arg2", (xmlChar*)$3) ;
	       $$ = (xmlChar*) PrecNode ;
             } ;

Desc: DESC LPAREN
      LSQBK Features RSQBK
      RPAREN
             {
	       DBUGYACC(printf ("Propriete desc:%s %o\n",$4,$4););
	       xmlNodePtr DescNode = xmlNewChild (CurrentClass, NULL,
						  (xmlChar*)"description", NULL) ;
	       xmlAddChild (DescNode, (xmlNodePtr) $4) ;
	       $$ = (xmlChar*) DescNode ;
             } ;

Features: 
             {
	       xmlNodePtr FSNode = xmlNewNode (NULL, (xmlChar*)"fs") ;
	       DBUGYACC(printf ("Features : liste vide %o\n", FSNode););
	       $$ = (char*) FSNode ;
             }
     | F
             {
	       xmlNodePtr FNode = (xmlNodePtr) $1 ;
	       xmlNodePtr FSNode = xmlNewNode (NULL, (xmlChar*)"fs") ;
	       DBUGYACC(printf ("Features : F %o %o\n", $1, FSNode););
	       xmlAddChild (FSNode, FNode) ;
	       $$ = (char*) FSNode ;
             }
     | Features COMMA F
             {
	       xmlNodePtr FNode = (xmlNodePtr) $3 ;
	       xmlNodePtr FSNode = (xmlNodePtr) $1 ;
	       DBUGYACC(printf ("Features : F FS %o %o %o\n", $3, $2, FSNode););
	       xmlAddChild (FSNode, FNode) ;
	       $$ = (char*) FSNode ;
             } ;

F: IDENT COLON Value
             {
	       xmlNodePtr FNode = xmlNewNode (NULL, (xmlChar*)"f") ;
	       DBUGYACC(printf ("Feature %s %o %o\n", $1, $3, FNode););
	       xmlSetProp (FNode, (xmlChar*)"name", (xmlChar*)$1) ;
	       xmlAddChild (FNode, (xmlNodePtr)$3) ;
	       $$ = (char*) FNode ;
             } ;


Guard: Idname IMP EqAlt
             { 
               xmlNodePtr GuardNode = (xmlNodePtr) $3;
               xmlSetProp (GuardNode, (xmlChar*)"rel", (xmlChar *) "+");
               xmlSetProp (GuardNode, (xmlChar*)"name", (xmlChar *) $1);
               xmlAddChild (CurrentClass, GuardNode) ;
	       DBUGYACC(printf("Guard: GuardNode=%o %s\n",GuardNode, $1););
	       $$ = (char*) GuardNode ;
             } 
      | NEG Idname IMP EqAlt
             { 
               xmlNodePtr GuardNode = (xmlNodePtr) $4;
               xmlSetProp (GuardNode, (xmlChar*)"rel", (xmlChar *) "-");
               xmlSetProp (GuardNode, (xmlChar*)"name", (xmlChar *) $2);
               xmlAddChild (CurrentClass, GuardNode) ;
	       DBUGYACC(printf("Guard: GuardNode=%o ~%s\n",GuardNode,$2););
	       $$ = (char*) GuardNode ;
             }
      | Idname PLUS EqAlt
            {
               xmlNodePtr GuardNode = (xmlNodePtr) $3;
               xmlSetProp (GuardNode, (xmlChar*)"rel", (xmlChar *) "+");
               xmlSetProp (GuardNode, (xmlChar*)"name", (xmlChar *) $1);
               xmlAddChild (CurrentClass, GuardNode) ;
               xmlNodePtr FailNode = xmlNewNode(NULL, (xmlChar*)"guard");
               xmlSetProp (FailNode, (xmlChar*)"rel", (xmlChar *) "-");
               xmlSetProp (FailNode, (xmlChar*)"name", (xmlChar *) $1);
               xmlAddChild (CurrentClass, FailNode) ;
	       DBUGYACC(printf("Guard: GuardNode=%o %s\n",GuardNode, $1););
	       $$ = (char*) GuardNode ;
             }
      | Idname XPLUS EqAlt
            {
               xmlNodePtr GuardNode = (xmlNodePtr) $3;
               xmlSetProp (GuardNode, (xmlChar*)"rel", (xmlChar *) "++");
               xmlSetProp (GuardNode, (xmlChar*)"name", (xmlChar *) $1);
               xmlAddChild (CurrentClass, GuardNode) ;
               xmlNodePtr FailNode = xmlNewNode(NULL, (xmlChar*)"guard");
               xmlSetProp (FailNode, (xmlChar*)"rel", (xmlChar *) "-");
               xmlSetProp (FailNode, (xmlChar*)"name", (xmlChar *) $1);
               xmlAddChild (CurrentClass, FailNode) ;
	       DBUGYACC(printf("Guard: GuardNode=%o %s\n",GuardNode, $1););
	       $$ = (char*) GuardNode ;
             }
      ;     

EqAlt: EqSeq
             { 
                 xmlNodePtr GuardNode = xmlNewNode (NULL, (xmlChar*)"guard");
                 xmlAddChild (GuardNode,(xmlNodePtr)$1);
		 DBUGYACC(printf("EqAlt: GuardNode=%o\n",GuardNode););
                 $$ = (char *)GuardNode;
             }
     | EqAlt ALT EqSeq
             {    xmlNodePtr GuardNode = (xmlNodePtr) $1;
	          xmlAddChild (GuardNode,(xmlNodePtr) $3);
		  DBUGYACC(printf("EqAlt: GuardNode=%o\n",GuardNode););
		  $$ = $1;
	     } ;

EqSeq: EqElem { $$ = $1; }
     | EqSeq1 COMMA EqElem {
         xmlNodePtr SeqGuardNode = (xmlNodePtr) $1;
         xmlAddChild (SeqGuardNode,(xmlNodePtr) $3);
	 DBUGYACC(printf("EqSeq: SeqGuardNode=%o\n",SeqGuardNode););
         $$ = $1;
     };

EqSeq1: EqElem {
       xmlNodePtr SeqGuardNode = xmlNewNode (NULL, (xmlChar*)"and");
       xmlAddChild (SeqGuardNode,(xmlNodePtr)$1);
       DBUGYACC(printf ("EqSeq1 (1): SeqGuardNode=%o %o\n", SeqGuardNode));
       $$ = (char *)SeqGuardNode;
              }
       | EqSeq1 COMMA EqElem 
              {
                  xmlNodePtr SeqGuardNode = (xmlNodePtr) $1;
	          xmlAddChild (SeqGuardNode,(xmlNodePtr) $3);
		  DBUGYACC(printf ("EqSeq1: SeqGuardNode=%o %o\n", SeqGuardNode, $3));
		  $$ = $1;
              };

EqElem: EqTerm EQUAL EqTerm
             { 
               xmlNodePtr EqNode = xmlNewNode (NULL, (xmlChar*)"equation");
               xmlAddChild (EqNode, (xmlNodePtr) $1);
               xmlAddChild (EqNode, (xmlNodePtr) $3);
               $$ = (char *)EqNode;
	       DBUGYACC(printf ("EqElem: EqNode=%o\n", EqNode));
             }
      | EqTerm XEQUAL EqTerm
             { 
               xmlNodePtr EqNode = xmlNewNode (NULL, (xmlChar*)"equation");
               xmlSetProp (EqNode, (xmlChar*)"default", (xmlChar *) "+");
               xmlAddChild (EqNode, (xmlNodePtr) $1);
               xmlAddChild (EqNode, (xmlNodePtr) $3);
               $$ = (char *)EqNode;
	       DBUGYACC(printf ("EqElem: EqNode=%o\n", EqNode));
             }
      ;

Atomic: BANG EqTerm
             { 
               xmlNodePtr AtomicNode = xmlNewNode (NULL, (xmlChar*)"atomic");
               xmlAddChild (AtomicNode, (xmlNodePtr) $2);
               xmlAddChild (CurrentClass, AtomicNode) ;
               $$ = (char *)AtomicNode;
             } ;

Equation: EqTerm EQUAL EqTerm
             {
	       xmlNodePtr EquationNode = xmlNewNode (NULL, (xmlChar*)"equation") ;
	       xmlNodePtr Root1 = (xmlNodePtr)$1 ;
	       xmlNodePtr Root2 = (xmlNodePtr)$3 ;
	       xmlAddChild (EquationNode, Root1) ;
	       xmlAddChild (EquationNode, Root2) ;
	       xmlAddChild (CurrentClass, EquationNode) ;
	       DBUGYACC(printf ("Propriete Equation : Equation_node=%o Root1=%o Root2=%o\n", EquationNode, $1, $3););
               $$ = (char *)EquationNode ;
             } ;

EqTerm: RootOfPath Path
             {
	       xmlNodePtr RootNode = (xmlNodePtr)$1 ;
	       xmlAddChild (RootNode, (xmlNodePtr)$2) ;
	       DBUGYACC(printf ("EqTerm: RootNode:%o Path:%o\n",RootNode,$2););
	       $$ = (char*) RootNode ;
             }
    | VALUE LPAREN Value RPAREN
             {
	       xmlNodePtr Node = xmlNewNode (NULL, (xmlChar*)"value") ;
	       xmlAddChild (Node, (xmlNodePtr)$3) ;
	       DBUGYACC(printf ("EqTerm: Node:%o Value:%o\n",Node,$3););
	       $$ = (char*) Node ;
             } ;

Varname: VAR Idname
           {
	       xmlNodePtr VarNode = xmlNewNode (NULL, (xmlChar*)"var") ;
	       xmlSetProp (VarNode, (xmlChar*)"name", (xmlChar*)$2) ;
	       $$ = (char*) VarNode ;
           } ;

AtomAlts: Atom
             {
                 xmlNodePtr AltNode = xmlNewNode (NULL, (xmlChar*)"vAlt") ;
                 xmlAddChild (AltNode, (xmlNodePtr) $1) ;
                 $$ = (char*) AltNode ;
	     }
        | Atom ALT AtomAlts
             {
	       xmlNodePtr AltNode = (xmlNodePtr)$3 ;
	       xmlNodePtr FirstAltNode = AltNode->xmlChildrenNode ;
	       xmlAddPrevSibling (FirstAltNode, (xmlNodePtr)$1) ;
	       $$ = (char*) AltNode ;
	     } ;

Atom: IDENT {
	       xmlNodePtr VarNode = xmlNewNode (NULL, (xmlChar*)"symbol") ;
	       xmlSetProp (VarNode, (xmlChar*)"value", (xmlChar*)$1) ;
	       $$ = (char*) VarNode ;
	     }

     | STAR
             {
	       xmlNodePtr VarNode = xmlNewNode (NULL, (xmlChar*)"symbol") ;
	       xmlSetProp (VarNode, (xmlChar*)"value", (xmlChar*)"*") ;
	       $$ = (char*) VarNode ;
	     }
     | PLUS
             {
	       $$ = (char*) xmlNewNode (NULL, (xmlChar*)"plus") ;
	     }
     | MINUS
             {
	       $$ = (char*) xmlNewNode (NULL, (xmlChar*)"minus") ;
	     } ;

NValue:  Atom { $$ = $1; }
     | Atom ALT AtomAlts
             {
	       xmlNodePtr AltNode = (xmlNodePtr)$3 ;
	       xmlNodePtr FirstAltNode = AltNode->xmlChildrenNode ;
	       xmlAddPrevSibling (FirstAltNode, (xmlNodePtr) $1) ;
	       $$ = (char*) AltNode ;
	     } ;

Value: NValue { $$ = $1; }
     | NEG NValue
             {
	       xmlNodePtr NotNode = xmlNewNode (NULL, (xmlChar*)"not") ;
               xmlAddChild(NotNode, (xmlNodePtr)$2);
	       $$ = (char*) NotNode ;
	     }
     | Varname
             {
	       $$ = $1;
	     }
     | Varname JOIN Value
             {
                 xmlNodePtr VarNode = (xmlNodePtr) $1;
                 xmlAddChild (VarNode, (xmlNodePtr)$3) ;
                 $$ = (char*) VarNode ;
	     }
     | MACRONAME
            {
	       xmlNodePtr FNode = xmlNewNode (NULL, (xmlChar*)"fs") ;
	       xmlSetProp (FNode, (xmlChar*)"idref", (xmlChar*)$1) ;
	       $$ = (char*) FNode ;
             }
     | LSQBK Features RSQBK
             {
	       $$ = (char*) $2 ;
             } ;

Node: VNode
     | OPT VNode
            {
	      xmlNodePtr Node = (xmlNodePtr) $2 ;
	      xmlSetProp (Node, (xmlChar*)"optional", (xmlChar*)"yes") ;
	      $$ = (char*) Node ;
	    } ;

VNode: RawNode
            {
	      xmlNodePtr Node = (xmlNodePtr) $1 ;
	      xmlSetProp (Node, (xmlChar*)"virtual", (xmlChar*)"yes") ;
	      $$ = (char*) Node ;
	    }
     | NODE RawNode
            {
	      xmlNodePtr Node = (xmlNodePtr) $2 ;
	      xmlSetProp (Node, (xmlChar*)"virtual", (xmlChar*)"no") ;
	      $$ = (char *) Node ;
	    } ;

RawNode: Idname COLON Value
           {
	       DBUGYACC (printf ("Node %s\n",$1););
	       xmlNodePtr Node = xmlNewNode (NULL, (xmlChar*)"node") ;
	       xmlSetProp (Node, (xmlChar*)"name", (xmlChar*)$1) ;
//               xmlSetProp (Node, (xmlChar*)"virtual", (xmlChar*)"yes") ;
	       xmlAddChild (Node, (xmlNodePtr)$3) ;
	       xmlAddChild (CurrentClass, Node) ;
               $$= (char *) Node;
           } ;

Path: 
             {
	       xmlNodePtr PathNode = NULL ;
	       $$ = (char*) PathNode ;
	     }
    | DOT IDENT Path 
             {
	       xmlNodePtr PathNode = xmlNewNode (NULL, (xmlChar*)"f") ;
	       xmlSetProp (PathNode, (xmlChar*)"name", (xmlChar*)$2) ;
	       if ($3 != NULL) {
		 xmlAddChild (PathNode, (xmlNodePtr)$3) ;
	       }
	       $$ = (char*) PathNode ;
	     } 
    | DOT MACRONAME Path
             {
	       xmlNodePtr PathNode = xmlNewNode (NULL, (xmlChar*)"path") ;
	       xmlSetProp (PathNode, (xmlChar*)"idref", (xmlChar*)$2) ;
	       if ($3 != NULL) {
		 xmlAddChild (PathNode, (xmlNodePtr)$3) ;
	       }
	       new_path_macro_symb ($2, 0, yylineno) ;
	       $$ = (char*) PathNode ;
	     } ;

RootOfPath : DESC
             {
	       xmlNodePtr Node = xmlNewNode (NULL, (xmlChar*)"description") ;
	       $$ = (char*) Node ;
	     }
    | FATHER LPAREN Idname RPAREN
             {
	       xmlNodePtr Node = xmlNewNode (NULL, (xmlChar*)"father") ;
	       xmlSetProp (Node, (xmlChar*)"name", (xmlChar*)$3) ;
	       $$ = (char*) Node ;
	     }
    | NODE LPAREN Idname RPAREN
             {
	       xmlNodePtr Node = xmlNewNode (NULL, (xmlChar*)"node") ;
	       xmlSetProp (Node, (xmlChar*)"name", (xmlChar*)$3) ;
	       $$ = (char*) Node ;
	     }
   | Varname
             { $$ = $1; } ;

Needs: MINUS Idname
             {
	       DBUGYACC(printf ("Propriete Needs:%s\n",$1););
	       xmlNodePtr Node = xmlNewChild (CurrentClass, NULL, (xmlChar*)"needs", NULL) ;
	       xmlSetProp (Node, (xmlChar*)"name", (xmlChar*)$2) ;
	       $$ = (char*) Node ;
             } ;

Provides: PLUS Idname
             {
	       DBUGYACC(printf ("Propriete Provides:%s\n",$1););
	       xmlNodePtr Node = xmlNewChild (CurrentClass, NULL, (xmlChar*)"provides", NULL) ;
	       xmlSetProp (Node, (xmlChar*)"name", (xmlChar*)$2) ;
	       $$ = (char*) Node ;
             } ;

Idname: IDENT
             {
	       $$ = (char*) $1 ;
             }
    | Idname DOUBLECOLON IDENT
             {
	       char * str1 = (char*) $1 ;
	       char * str2 = (char*) $3 ;
	       char * newstr = (char*) calloc (strlen(str1) + 2 + strlen(str2),
					       sizeof (char)) ;
	       sprintf (newstr, "%s:%s", str1, str2) ;
	       $$ = newstr ;
             } ;

Doc: { $$ = NULL ; }
| DOCUMENTATION Doc
             {
	       DBUGYACC(printf ("Doc:\t%s\n", $1););
	       char * str1 = $1; char * str2 = $2 ;
	       $$ = cp_str (str1,str2) ;
             } ;
